'''
Iterators
=========

.. autoclass:: APIIterator
    :members:
    :private-members:
'''


class APIIterator(object):
    '''
    The API iterator provides a scalable way to work through result sets of any
    size.  The iterator will walk through each page of data, returning one
    record at a time.  If it reaches the end of a page of records, then it will
    request the next page of information and then continue to return records
    from the next page (and the next, and the next) until the counter reaches
    the total number of records that the API has reported.

    Note that this Iterator is used as a base model for all of the iterators,
    and while the mechanics of each iterator may vary, they should all behave
    to the user in a similar manner.

    Attributes:
        _api (restfly.session.APISession):
            The APISession object that will be used for querying for the
            data.
        count (int):
            The current number of records that have been returned
        max_items (int):
            The maximum number of items to return before stopping iteration.
        max_pages (int):
            The maximum number of pages to request before throwing stopping
            iteration.
        num_pages (int):
            The number of pages that have been requested.
        page (list):
            The current page of data being walked through.  pages will be
            cycled through as the iterator requests more information from the
            API.
        page_count (int): The number of record returned from the current page.
        total (int):
            The total number of records that exist for the current request.
    '''
    count = 0
    page_count = 0
    num_pages = 0
    max_pages = None
    max_items = None
    total = None
    page = []
    _api = None

    def __init__(self, api, **kw):
        '''
        Args:
            api (restfly.session.APISession):
                The APISession object to use for this iterator.
            **kw (dict):
                The various attributes to add and or overload the iterator with.

        Examples:
            >>> i = APIIterator(api, max_pages=1, max_items=100)
        '''
        self._api = api
        self.__dict__.update(kw)

    def _get_page(self):
        '''
        A method to be overloaded in order to instruct the iterator how to
        retrieve the next page of data.

        Examples:
            >>> class ExampleIterator(APIIterator):
            ...    def _get_page(self):
            ...        self.total = 100
            ...        self.page = [{'id': i + self._offset} for i in range(10)]
            ...        self._offset += self._limit
        '''
        pass

    def get(self, index, default=None):
        '''
        Retrieves an item from the the current page based off the index provided.

        Args:
            index (int): The index of the item to retrieve.
            default (obj): The returned object if the item does not exist.

        Examples:
            >>> a = APIIterator()
            >>> a.get(2)
            None
        '''
        try:
            return self.page[int(index)]
        except IndexError:
            return default

    def __getitem__(self, key):
        return self.page[key]

    def __iter__(self):
        return self

    def __next__(self):
        return self.next()

    def next(self):
        '''
        Ask for the next record
        '''
        # If there are no more records to return, then we should raise a
        # StopIteration exception to break the iterator out.
        if ((self.total and self.count + 1 > self.total)
          or (self.max_items and self.count + 1 > self.max_items)
          or (self.max_pages and self.num_pages > self.max_pages)):
            raise StopIteration()

        # If we have worked through the current page of records and we still
        # haven't hit to the total number of available records, then we should
        # query the next page of records.
        if (self.page_count >= len(self.page)
          and (not self.total or self.count + 1 <= self.total)):

            # If the number of pages requested reaches the total number of pages
            # that should be requested, then stop iteration.
            if self.max_pages and self.num_pages + 1 > self.max_pages:
                raise StopIteration()

            # Perform the _get_page call.
            self.page_count = 0
            self.num_pages += 1
            self._get_page()

            # If the length of the page is 0, then we don't have anything
            # further to do and should stop iteration.
            if len(self.page) == 0:
                raise StopIteration()

        # Get the relevant record, increment the counters, and return the
        # record.
        self.count += 1
        self.page_count += 1
        item = self[self.page_count - 1]
        return item


'''
Sessions
========

.. autoclass:: APISession
    :members:
    :private-members:
'''
import requests, sys, time, warnings, os
from .errors import *
from . import __version__

try:
    from urlparse import urlparse
except:
    from urllib.parse import urlparse

class APISession(object):
    '''
    The APISession class is the base model for APISessions for different
    products and applications.  This is the model that the APIEndpoints
    will be grafted onto and supports some basic wrapping of standard HTTP
    methods on it's own.

    Attributes:
        _build (str):
            The build number/version of the integration.
        _backoff (float):
            The default backoff timer to use when retrying.  The value is either
            a float or integer denoting the number of seconds to delay before
            the next retry attempt.  The number will be multiplied by the number
            of retries attempted.
        _error_map (dict):
            The error mapping detailing what HTTP response code should throw
            what kind of error.  Specifically meant to be overloadable based on
            need.
        _lib_name (str):
            The name of the library.
        _lib_version (str):
            The version of the library.
        _product (str):
            The product name for the integration.
        _proxies (dict):
            A dictionary detailing what proxy should be used for what transport
            protocol.  This value will be passed to the session object after it
            has been either attached or created.  For details on the structure
            of this dictionary, consult the
            :requests:`proxies section of the Requests documentation.<user/advanced/#proxies>`
        _restricted_paths (list):
            A list of paths (not complete URIs) that if seen be the
            :obj:`_request` method will not pass the query params or the request
            body into the logging facility.  This should generally be used for
            paths that are sensitive in nature (such as logins).
        _retries (int):
            The number of retries to make before failing a request.  The
            default is 3.
        _session (requests.Session):
            Provide a pre-built session instead of creating a requests session
            at instantiation.
        _ssl_verify (bool):
            Should SSL verification be performed?  If not, then inform requests
            that we don't want to use SSL verification and suppress the SSL
            certificate warnings.
        _url (str):
            The base URL path to use.  This should generally be a string value
            denoting the first half of the URI.  For example,
            ``https://httpbin.org`` or ``https://example.api.site/api/2``.  The
            :obj:`_request` method will join this string with the incoming path
            to construct the complete URI.  Note that the two strings will be
            joined with a backslash ``/``.
        _vendor (str):
            The vendor name for the integration.
    '''
    _url = None
    _retries = 3
    _backoff = 1
    _proxies = None
    _ssl_verify = True
    _lib_name = 'Restfly'
    _lib_version = __version__
    _restricted_paths = list()
    _vendor = 'unknown'
    _product = 'unknown'
    _build = __version__
    _adaptor = None
    _error_map = {
        400: BadRequestError,
        401: UnauthorizedError,
        403: ForbiddenError,
        404: NotFoundError,
        405: InvalidMethodError,
        406: NotAcceptableError,
        407: ProxyAuthenticationError,
        408: RequestTimeoutError,
        409: RequestConflictError,
        410: NoLongerExistsError,
        411: LengthRequiredError,
        412: PreconditionFailedError,
        413: PayloadTooLargeError,
        414: URITooLongError,
        415: UnsupportedMediaTypeError,
        416: RangeNotSatisfiableError,
        417: ExpectationFailedError,
        418: TeapotResponseError,
        420: TooManyRequestsError,
        421: MisdirectRequestError,
        425: TooEarlyError,
        426: UpgradeRequiredError,
        428: PreconditionRequiredError,
        429: TooManyRequestsError,
        431: RequestHeaderFieldsTooLargeError,
        451: UnavailableForLegalReasonsError,
        500: ServerError,
        501: MethodNotImplementedError,
        502: BadGatewayError,
        503: ServiceUnavailableError,
        504: GatewayTimeoutError,
        510: NotExtendedError,
        511: NetworkAuthenticationRequiredError,
    }

    def __init__(self, **kwargs):
        '''
        APISession initialization

        Args:
            adaptor (Object, optional):
                A Requests Session adaptor to bind to the session object.
            backoff (float, optional):
                If a 429 response is returned, how much do we want to backoff
                if the response didn't send a Retry-After header.
            build (str, optional):
                The build number to put into the User-Agent string.
            product (str, optional):
                The product name to put into the User-Agent string.
            proxies (dict, optional):
                A dictionary detailing what proxy should be used for what
                transport protocol.  This value will be passed to the session
                object after it has been either attached or created.  For
                details on the structure of this dictionary, consult the
                :requests:`proxies <user/advanced/#proxies>` section of the
                Requests documentation.
            retries (int, optional):
                The number of retries to make before failing a request.  The
                default is 3.
            session (requests.Session, optional):
                Provide a pre-built session instead of creating a requests
                session at instantiation.
            ssl_verify (bool, optional):
                If SSL Verification needs to be disabled (for example when using
                a self-signed certificate), then this parameter should be set to
                ``False`` to disable verification and mask the Certificate
                warnings.
            url (str, optional):
                The base URL that the paths will be appended onto.
            vendor (str, optional):
                The vendor name to put into the User-Agent string.
        '''
        # Assign the kw arguments to the private attributes.
        self._url = kwargs.get('url', self._url)
        self._retries = int(kwargs.get('retries', self._retries))
        self._backoff = float(kwargs.get('backoff', self._backoff))
        self._proxies = kwargs.get('proxies', self._proxies)
        self._ssl_verify = kwargs.get('ssl_verify', self._ssl_verify)
        self._adaptor = kwargs.get('adaptor', self._adaptor)
        self._vendor = kwargs.get('vendor', self._vendor)
        self._product = kwargs.get('product', self._product)
        self._build = kwargs.get('build', self._build)

        # Create the logging facility
        self._log = logging.getLogger('{}.{}'.format(
            self.__module__, self.__class__.__name__))

        # Initiate the session builder.
        self._build_session(**kwargs)

    def _build_session(self, **kwargs):
        '''
        The session builder.  User-agent strings, cookies, headers, etc that
        should persist for the session should be initiated here.  The session
        builder is called as part of the APISession constructor.

        Args:
            session (requests.Session, optional):
                If a session object was passed to the constructor, then this
                would contain a session, otherwise a new one is created.

        Returns:
            :obj:`None`

        Examples:
            Extending the session builder to use basic auth:

            >>> class ExampleAPI(APISession):
            ...     def _build_session(self, session=None):
            ...         super(APISession, self)._build_session(**kwargs)
            ...         self._session.auth = (self._username, self._password)
        '''
        uname = os.uname()
        # link up the session to either the one passed or create a new session.
        self._session = kwargs.get('session', requests.Session())

        # If proxy support is needed, update the proxies in the session.
        if self._proxies:
            self._session.proxies.update(self._proxies)

        # If the SSL verification is disabled then we will need to disable
        # verification in the requests session and we also want to mask the
        # certificate warnings.
        if not self._ssl_verify:
            self._session.verify = self._ssl_verify
            warnings.filterwarnings('ignore', 'Unverified HTTPS request')

        # Update the User-Agent string with the information necessary.
        self._session.headers.update({
            'User-Agent': ' '.join([
                'Integration/1.0 ({}; {}; Build/{})'.format(

                    # The vendor name for the integration
                    self._vendor,

                    # The product name of the integration
                    self._product,

                    # The build of the integration
                    self._build
                ),
                '{}/{} (Restfly/{}; Python/{}; {}/{})'.format(

                    # The name of the library being used
                    self._lib_name,

                    # The version of the library being used
                    self._lib_version,

                    # The version of Restfly
                    __version__,

                    # The python version string
                    '.'.join([str(i) for i in sys.version_info][0:3]),

                    # The source OS
                    uname[0],

                    # The source Arch
                    uname[-1]
                ),
            ])
        })

    def _resp_error_check(self, response, **kwargs): #stub
        '''
        If there is a need for additional error checking (for example within the
        JSON response) then overload this method with the necessary checking.

        Args:
            response (request.Response):
                The response object.
            **kwargs (dict):
                The request keyword arguments.

        Returns:
            :obj:`requests.Response`:
                The response object.
        '''
        return response

    def _retry_request(self, response, retries, **kwargs): #stub
        '''
        A method to be overloaded to return any modifications to the request
        upon retries.  By default just passes back what was send in the same
        order.

        Args:
            response (request.Response):
                The response object
            retries (int):
                The number of retries that have been performed.
            **kwargs (dict):
                The keyword arguments that were passed to the request.

        Returns:
            :obj:`dict`:
                The keyword arguments
        '''
        return kwargs

    def _request(self, method, path, **kwargs):
        '''
        The requests session base request method.  This is considered internal
        as it's generally recommended to use the bespoke methods for each HTTP
        method.

        Args:
            method (str):
                The HTTP method
            path (str):
                The URI path to append to the base path.
            **kwargs (dict):
                The keyword arguments to pass to the requests lib.
            retry_on (list, optional):
                A list of numeric response status codes to attempt retry on.
                This behavior is additive to the retry parameter in the
                exceptions.

        Returns:
            :obj:`requests.Response`:
                The response object from the requests lib.

        Examples:
            >>> api = APISession()
            >>> resp = api._request('GET', '/')
        '''
        err = None
        retries = 0

        # If retry_on is specified, then we will populate the retry_codes
        # variable with a list of numeric status codes to additionally retry on.
        # This is helpful if the API in question doesn't always behave in a
        # consistent manner.
        retry_codes = kwargs.pop('retry_on', list())

        # While the number of retries is less than the retry limit, loop.  As we
        # will be returning from within the loop if we receive a successful
        # response or a non-retryable error, the loop should only be handling
        # the retries themselves.
        while retries <= self._retries:
            # Check to see if the path is a relative path or a
            if urlparse(path).netloc != '':
                uri = path
            else:
                uri = '{}/{}'.format(self._url, path)

            if (('params' in kwargs and kwargs['params'])
              or ('json' in kwargs and kwargs['json'])):
                if path not in self._restricted_paths:
                    # If the path is not one of the paths that would contain
                    # sensitive data (such as login information) then pass the
                    # log on unredacted.
                    self._log.debug('uri={}, query={}, body={}'.format(
                        uri, kwargs.get('params', {}), kwargs.get('json', {})))
                else:
                    # The path was a restricted path, generate the log, however
                    # redact the information.
                    self._log.debug('uri={}, query={}, body={}'.format(
                        uri, 'REDACTED', 'REDACTED'))

            # Make the call to the API and pull the status code.
            resp = self._session.request(method, uri, **kwargs)
            status = resp.status_code

            if status in self._error_map.keys():
                # If a status code that we know about has returned, then we will
                # want to raise the appropriate Error.
                err = self._error_map[status]
                if err.retryable or status in retry_codes:
                    # If the APIError fetched is retryable, we will want to
                    # attempt to retry our call.  If we see the "Retry-After"
                    # header, then we will respect that.  If no "Retry-After"
                    # header exists, then we will use the _backoff attribute to
                    # build a back-off timer based on the number of retries we
                    # have already performed.
                    retries += 1
                    time.sleep(resp.headers.get(
                        'retry-after', retries * self._backoff))

                    # The need to potentially modify the request for subsequent
                    # calls is the whole reason that we aren't using the default
                    # Retry logic that urllib3 supports.
                    kwargs = self._retry_request(resp, retries, **kwargs)
                    continue
                else:
                    raise err(resp, retries=retries)

            elif status >= 200 and status <= 299:
                # As everything looks ok, lets pass the response on to the error
                # checker and then return the response.
                return self._resp_error_check(resp, **kwargs)

            else:
                # If all else fails, raise an error stating that we don't even
                # know whats happening.
                raise APIError(resp, retries=retries)
        raise err(resp, retries=retries)

    def get(self, path, **kwargs):
        '''
        Initiates an HTTP GET request using the specified path.  Refer to
        :obj:`requests.request` for more detailed information on what
        keyword arguments can be passed:

        Args:
            path (str):
                The path to be appended onto the base URL for the request.
            **kwargs (dict):
                Keyword arguments to be passed to the Requests library.

        Returns:
            :obj:`requests.Response`

        Examples:
            >>> api = APISession()
            >>> resp = api.get('/')
        '''
        return self._request('GET', path, **kwargs)

    def post(self, path, **kwargs):
        '''
        Initiates an HTTP POST request using the specified path.  Refer to the
        :obj:`requests.request` for more detailed information on what
        keyword arguments can be passed:

        Args:
            path (str):
                The path to be appented onto the base URL for the request.
            **kwargs (dict):
                Keyword arguments to be passed to the Requests library.

        Returns:
            :obj:`requests.Response`

        Examples:
            >>> api = APISession()
            >>> resp = api.post('/')
        '''
        return self._request('POST', path, **kwargs)

    def put(self, path, **kwargs):
        '''
        Initiates an HTTP PUT request using the specified path.  Refer to the
        :obj:`requests.request` for more detailed information on what
        keyword arguments can be passed:

        Args:
            path (str):
                The path to be appended onto the base URL for the request.
            **kwargs (dict):
                Keyword arguments to be passed to the Requests library.

        Returns:
            :obj:`requests.Response`

        Examples:
            >>> api = APISession()
            >>> resp = api.put('/')
        '''
        return self._request('PUT', path, **kwargs)

    def patch(self, path, **kwargs):
        '''
        Initiates an HTTP PATCH request using the specified path.  Refer to the
        :obj:`requests.request` for more detailed information on what
        keyword arguments can be passed:

        Args:
            path (str):
                The path to be appended onto the base URL for the request.
            **kwargs (dict):
                Keyword arguments to be passed to the Requests library.

        Returns:
            :obj:`requests.Response`

        Examples:
            >>> api = APISession()
            >>> resp = api.patch('/')
        '''
        return self._request('PATCH', path, **kwargs)

    def delete(self, path, **kwargs):
        '''
        Initiates an HTTP DELETE request using the specified path.  Refer to the
        :obj:`requests.request` for more detailed information on what
        keyword arguments can be passed:

        Args:
            path (str):
                The path to be appended onto the base URL for the request.
            **kwargs (dict):
                Keyword arguments to be passed to the Requests library.

        Returns:
            :obj:`requests.Response`

        Examples:
            >>> api = APISession()
            >>> resp = api.delete('/')
        '''
        return self._request('DELETE', path, **kwargs)

    def head(self, path, **kwargs):
        '''
        Initiates an HTTP HEAD request using the specified path.  Refer to the
        :obj:`requests.request` for more detailed information on what
        keyword arguments can be passed:

        Args:
            path (str):
                The path to be appended onto the base URL for the request.
            **kwargs (dict):
                Keyword arguments to be passed to the Requests library.

        Returns:
            :obj:`requests.Response`

        Examples:
            >>> api = APISession()
            >>> resp = api.head('/')
        '''
        return self._request('HEAD', path, **kwargs)




'''
Errors
======

.. autoclass:: RestflyException
.. autoclass:: UnexpectedValueError
.. autoclass:: APIError
.. autoclass:: BadRequestError
.. autoclass:: UnauthorizedError
.. autoclass:: ForbiddenError
.. autoclass:: NotFoundError
.. autoclass:: InvalidMethodError
.. autoclass:: NotAcceptableError
.. autoclass:: ProxyAuthenticationError
.. autoclass:: RequestTimeoutError
.. autoclass:: RequestConflictError
.. autoclass:: NoLongerExistsError
.. autoclass:: LengthRequiredError
.. autoclass:: PreconditionFailedError
.. autoclass:: PayloadTooLargeError
.. autoclass:: URITooLongError
.. autoclass:: UnsupportedMediaTypeError
.. autoclass:: RangeNotSatisfiableError
.. autoclass:: ExpectationFailedError
.. autoclass:: TeapotResponseError
.. autoclass:: MisdirectRequestError
.. autoclass:: TooEarlyError
.. autoclass:: UpgradeRequiredError
.. autoclass:: PreconditionRequiredError
.. autoclass:: TooManyRequestsError
.. autoclass:: RequestHeaderFieldsTooLargeError
.. autoclass:: UnavailableForLegalReasonsError
.. autoclass:: ServerError
.. autoclass:: MethodNotImplementedError
.. autoclass:: BadGatewayError
.. autoclass:: ServiceUnavailableError
.. autoclass:: GatewayTimeoutError
.. autoclass:: NotExtendedError
.. autoclass:: NetworkAuthenticationRequiredError
'''
import logging

class RestflyException(Exception):
    '''
    Base exception class that sets up logging and handles some basic scaffolding
    for all other exception classes.  This exception should never be directly
    seen.
    '''
    def __init__(self, msg):
        self._log = logging.getLogger('{}.{}'.format(
            self.__module__, self.__class__.__name__))
        self.msg = str(msg)
        self._log.error(self.msg)

    def __str__(self):
        return self.msg

    def __repr__(self):
        return repr(self.__str__())


class UnexpectedValueError(RestflyException):
    '''
    An unexpected value error is thrown whenever the value specified for a
    parameter is outside the bounds of what is expected.  For example, if the
    parameter **a** is expected to have a value of 1, 2, or 3, and it is instead
    passed a value of 0, then it is an unexpected value, and this Exception
    should be thrown by the package.
    '''
    pass


# The following Exception codes have been written using the following link as
# a baseline (and the ):
# https://en.wikipedia.org/wiki/List_of_HTTP_status_codes


class APIError(RestflyException):
    '''
    The APIError Exception is a generic Exception for handling responses from
    the API that aren't whats expected.  The APIError Exception itself attempts
    to provide the developer with enough information around the response to
    ascertain what went wrong.

    Attributes:
        code (int):
            The HTTP response code from the offending response.
        response (request.Response):
            This is the Response object that had caused the Exception to fire.
    '''
    retryable = False
    retries = None

    def __init__(self, r, retries=None):
        self.response = r
        self.code = r.status_code
        self.retries = retries
        RestflyException.__init__(self, self.__str__())

    def __str__(self):
        return '[{}: {}] {} body={}'.format(
            str(self.code),
            str(self.response.request.method),
            str(self.response.request.url),
            str(self.response.content))


class BadRequestError(APIError):  # 400 Response
    '''
    The server cannot or will not process the request due to an apparent client
    error (e.g., malformed request syntax, size too large, invalid request
    message framing, or deceptive request routing).

    Attributes:
        code (int):
            The HTTP response code from the offending response.
        response (request.Response):
            This is the Response object that had caused the Exception to fire.
    '''
    pass


class UnauthorizedError(APIError):  # 401 Response
    '''
    Similar to 403 Forbidden, but specifically for use when authentication is
    required and has failed or has not yet been provided. The response must
    include a WWW-Authenticate header field containing a challenge applicable
    to the requested resource. See Basic access authentication and Digest access
    authentication. 401 semantically means "unauthenticated", i.e. the user
    does not have the necessary credentials.

    Attributes:
        code (int):
            The HTTP response code from the offending response.
        response (request.Response):
            This is the Response object that had caused the Exception to fire.
    '''


class ForbiddenError(APIError):  # 403 Response
    '''
    The request was valid, but the server is refusing action. The user might
    not have the necessary permissions for a resource, or may need an account
    of some sort.

    Attributes:
        code (int):
            The HTTP response code from the offending response.
        response (request.Response):
            This is the Response object that had caused the Exception to fire.
    '''
    pass


class NotFoundError(APIError):  # 404 Response
    '''
    The requested resource could not be found but may be available in the
    future. Subsequent requests by the client are permissible.

    Attributes:
        code (int):
            The HTTP response code from the offending response.
        response (request.Response):
            This is the Response object that had caused the Exception to fire.
    '''
    pass


class InvalidMethodError(APIError):  # 405 Response
    '''
    A request method is not supported for the requested resource; for example,
    a GET request on a form that requires data to be presented via POST, or a
    PUT request on a read-only resource.

    Attributes:
        code (int):
            The HTTP response code from the offending response.
        response (request.Response):
            This is the Response object that had caused the Exception to fire.
    '''
    pass


class NotAcceptableError(APIError):  # 406 Response
    '''
    The requested resource is only capable of generating content not
    acceptable according to the Accept headers sent in the request.

    Attributes:
        code (int):
            The HTTP response code from the offending response.
        response (request.Response):
            This is the Response object that had caused the Exception to fire.
    '''
    pass


class ProxyAuthenticationError(APIError):  # 407 Response
    '''
    The client must first authenticate itself with the proxy.

    Attributes:
        code (int):
            The HTTP response code from the offending response.
        response (request.Response):
            This is the Response object that had caused the Exception to fire.
    '''
    pass


class RequestTimeoutError(APIError):  # 408 Response
    '''
    The client did not produce a request within the time that the server was
    prepared to wait. The client MAY repeat the request without modifications
    at any later time.

    Attributes:
        code (int):
            The HTTP response code from the offending response.
        response (request.Response):
            This is the Response object that had caused the Exception to fire.
    '''
    pass


class RequestConflictError(APIError):  # 409 Response
    '''
    Indicates that the request could not be processed because of conflict in
    the current state of the resource, such as an edit conflict between
    multiple simultaneous updates.

    Attributes:
        code (int):
            The HTTP response code from the offending response.
        response (request.Response):
            This is the Response object that had caused the Exception to fire.
    '''
    pass


class NoLongerExistsError(APIError):  # 410 Response
    '''
    Indicates that the resource requested is no longer available and will not
    be available again. This should be used when a resource has been
    intentionally removed and the resource should be purged. Upon receiving a
    410 status code, the client should not request the resource in the future.
    Clients such as search engines should remove the resource from their
    indices. Most use cases do not require clients and search engines to purge
    the resource, and a "404 Not Found" may be used instead.

    Attributes:
        code (int):
            The HTTP response code from the offending response.
        response (request.Response):
            This is the Response object that had caused the Exception to fire.
    '''
    pass


class LengthRequiredError(APIError):  # 411 Response
    '''
    The request did not specify the length of its content, which is required by
    the requested resource.

    Attributes:
        code (int):
            The HTTP response code from the offending response.
        response (request.Response):
            This is the Response object that had caused the Exception to fire.
    '''
    pass


class PreconditionFailedError(APIError):  # 412 Response
    '''
    The server does not meet one of the preconditions that the requester put
    on the request.

    Attributes:
        code (int):
            The HTTP response code from the offending response.
        response (request.Response):
            This is the Response object that had caused the Exception to fire.
    '''
    pass


class PayloadTooLargeError(APIError):  # 413 Response
    '''
    The request is larger than the server is willing or able to process.

    Attributes:
        code (int):
            The HTTP response code from the offending response.
        response (request.Response):
            This is the Response object that had caused the Exception to fire.
    '''
    pass


class URITooLongError(APIError):  # 414 Response
    '''
    The URI provided was too long for the server to process. Often the result
    of too much data being encoded as a query-string of a GET request, in which
    case it should be converted to a POST request.

    Attributes:
        code (int):
            The HTTP response code from the offending response.
        response (request.Response):
            This is the Response object that had caused the Exception to fire.
    '''
    pass


class UnsupportedMediaTypeError(APIError): # 415 Response
    '''
    The request entity has a media type which the server or resource does not
    support. For example, the client uploads an image as image/svg+xml, but the
    server requires that images use a different format.

    Attributes:
        code (int):
            The HTTP response code from the offending response.
        response (request.Response):
            This is the Response object that had caused the Exception to fire.
    '''
    pass


class RangeNotSatisfiableError(APIError):  # 416 Response
    '''
    The client has asked for a portion of the file (byte serving), but the
    server cannot supply that portion. For example, if the client asked for a
    part of the file that lies beyond the end of the file.

    Attributes:
        code (int):
            The HTTP response code from the offending response.
        response (request.Response):
            This is the Response object that had caused the Exception to fire.
    '''
    pass


class ExpectationFailedError(APIError):  # 417 Response
    '''
    The server cannot meet the requirements of the Expect request-header field.

    Attributes:
        code (int):
            The HTTP response code from the offending response.
        response (request.Response):
            This is the Response object that had caused the Exception to fire.
    '''
    pass


class TeapotResponseError(APIError):  # 418 Response
    '''
    This code was defined in 1998 as one of the traditional IETF April Fools'
    jokes, in RFC 2324, Hyper Text Coffee Pot Control Protocol, and is not
    expected to be implemented by actual HTTP servers. The RFC specifies this
    code should be returned by teapots requested to brew coffee.

    Attributes:
        code (int):
            The HTTP response code from the offending response.
        response (request.Response):
            This is the Response object that had caused the Exception to fire.
    '''
    pass


class MisdirectRequestError(APIError):  # 421 Response
    '''
    The request was directed at a server that is not able to produce a response

    Attributes:
        code (int):
            The HTTP response code from the offending response.
        response (request.Response):
            This is the Response object that had caused the Exception to fire.
    '''
    pass


    '''
    The request failed because it depended on another request and that request
    failed.

    Attributes:
        code (int):
            The HTTP response code from the offending response.
        response (request.Response):
            This is the Response object that had caused the Exception to fire.
    '''
    pass


class TooEarlyError(APIError):  # 425 Response
    '''
    Indicates that the server is unwilling to risk processing a request that
    might be replayed.

    Attributes:
        code (int):
            The HTTP response code from the offending response.
        response (request.Response):
            This is the Response object that had caused the Exception to fire.
    '''
    pass


class UpgradeRequiredError(APIError):  # 426 Response
    '''
    The client should switch to a different protocol such as TLS/1.0, given in
    the Upgrade header field.

    Attributes:
        code (int):
            The HTTP response code from the offending response.
        response (request.Response):
            This is the Response object that had caused the Exception to fire.
    '''
    pass


class PreconditionRequiredError(APIError):  # 428 Response
    '''
    The origin server requires the request to be conditional. Intended to
    prevent the 'lost update' problem, where a client GETs a resource's state,
    modifies it, and PUTs it back to the server, when meanwhile a third party
    has modified the state on the server, leading to a conflict.

    Attributes:
        code (int):
            The HTTP response code from the offending response.
        response (request.Response):
            This is the Response object that had caused the Exception to fire.
    '''
    pass


class TooManyRequestsError(APIError):  # 420 & 429 Response
    '''
    The user has sent too many requests in a given amount of time. Intended for
    use with rate-limiting schemes.

    Attributes:
        code (int):
            The HTTP response code from the offending response.
        response (request.Response):
            This is the Response object that had caused the Exception to fire.
    '''
    retryable = True


class RequestHeaderFieldsTooLargeError(APIError):  # 431 Response
    '''
    The server is unwilling to process the request because either an individual
    header field, or all the header fields collectively, are too large.

    Attributes:
        code (int):
            The HTTP response code from the offending response.
        response (request.Response):
            This is the Response object that had caused the Exception to fire.
    '''
    pass


class UnavailableForLegalReasonsError(APIError):  # 451 Response
    '''
    A server operator has received a legal demand to deny access to a resource
    or to a set of resources that includes the requested resource.

    Attributes:
        code (int):
            The HTTP response code from the offending response.
        response (request.Response):
            This is the Response object that had caused the Exception to fire.
    '''
    pass


class ServerError(APIError):  # 500 Response
    '''
    A generic error message, given when an unexpected condition was encountered
    and no more specific message is suitable.

    Attributes:
        code (int):
            The HTTP response code from the offending response.
        response (request.Response):
            This is the Response object that had caused the Exception to fire.
    '''
    pass


class MethodNotImplementedError(APIError):  # 501 Response
    '''
    The server either does not recognize the request method, or it lacks the
    ability to fulfil the request. Usually this implies future availability.

    Attributes:
        code (int):
            The HTTP response code from the offending response.
        response (request.Response):
            This is the Response object that had caused the Exception to fire.
    '''
    retryable = True


class BadGatewayError(APIError):  # 502 Response
    '''
    The server was acting as a gateway or proxy and received an invalid
    response from the upstream server.

    Attributes:
        code (int):
            The HTTP response code from the offending response.
        response (request.Response):
            This is the Response object that had caused the Exception to fire.
    '''
    retryable = True


class ServiceUnavailableError(APIError):  # 503 Response
    '''
    The server cannot handle the request (because it is overloaded or down for
    maintenance). Generally, this is a temporary state.

    Attributes:
        code (int):
            The HTTP response code from the offending response.
        response (request.Response):
            This is the Response object that had caused the Exception to fire.
    '''
    retryable = True


class GatewayTimeoutError(APIError):  # 504 Response
    '''
    The server was acting as a gateway or proxy and did not receive a timely
    response from the upstream server.

    Attributes:
        code (int):
            The HTTP response code from the offending response.
        response (request.Response):
            This is the Response object that had caused the Exception to fire.
    '''
    retryable = True


class NotExtendedError(APIError):  # 510 Response
    '''
    Further extensions to the request are required for the server to fulfil it.

    Attributes:
        code (int):
            The HTTP response code from the offending response.
        response (request.Response):
            This is the Response object that had caused the Exception to fire.
    '''
    pass


class NetworkAuthenticationRequiredError(APIError):  # 511 Response
    '''
    The client needs to authenticate to gain network access. Intended for use
    by intercepting proxies used to control access to the network

    Attributes:
        code (int):
            The HTTP response code from the offending response.
        response (request.Response):
            This is the Response object that had caused the Exception to fire.
    '''
    pass


'''
Utils
=====

.. autofunction:: check
.. autofunction:: dict_merge
.. autofunction:: force_case
.. autofunction:: trunc

'''
from .errors import UnexpectedValueError
import re

def dict_merge(master, updates):
    '''
    Merge 2 dictionaries together  The updates dictionary will be merged into
    the master, adding/updating any values as needed.

    Args:
        master (dict):
            The master dictionary to be used as the base.
        updates (dict):
            The dictionary that will overload the values in the master.

    Returns:
        :obj:`dict`:
            The merged dictionary

    Examples:
        >>> a = {'one': 1, 'two': 2, 'three': {'four': 4}}
        >>> b = {'a': 'a', 'three': {'b': 'b'}}
        >>> dict_merge(a, b)
        {'a': 'a', 'one': 1, 'two': 2, 'three': {'b': b, 'four': 4}}
    '''
    for key in updates:
        if (key in master and isinstance(master[key], dict)
          and isinstance(updates[key], dict)):
            master[key] = dict_merge(master[key], updates[key])
        else:
            master[key] = updates[key]
    return master


def force_case(obj, case):
    '''
    A simple case enforcement function.

    Args:
        obj (Object): object to attempt to enforce the case upon.

    Returns:
        :obj:`obj`:
            The modified object

    Examples:
        A list of mixed types:

        >>> a = ['a', 'list', 'of', 'strings', 'with', 'a', 1]
        >>> force_Case(a, 'upper')
        ['A', 'LIST', 'OF', 'STRINGS', 'WITH', 'A', 1]

        A simple string:

        >>> force_case('This is a TEST', 'lower')
        'this is a test'

        A non-string item that'll pass through:

        >>> force_case(1, 'upper')
        1
    '''
    if case == 'lower':
        if isinstance(obj, list):
            return [i.lower() for i in obj if isinstance(i, str)]
        elif isinstance(obj, str):
            return obj.lower()

    elif case == 'upper':
        if isinstance(obj, list):
            return [i.upper() for i in obj if isinstance(i, str)]
        elif isinstance(obj, str):
            return obj.upper()

    return obj


def trunc(text, limit, suffix='...'):
    '''
    Truncates a string to a given number of characters.  If a string extends
    beyond the limit, then truncate and add an ellipses after the truncation.

    Args:
        text (str): The string to truncate
        limit (int): The maximum limit that the string can be.
        suffix (str):
            What suffix should be appended to the truncated string when we
            truncate?  If left unspecified, it will default to ``...``.


    Returns:
        :obj:`str`:
            The truncated string

    Examples:
        A simple truncation:

        >>> trunc('this is a test', 6)
        'thi...'

        Truncating with no suffix:

        >>> trunc('this is a test', 6, suffix=None)
        'this i'

        Truncating with a custom suffix:

        >>> trunc('this is a test', 6, suffix='->')
        'this->'
    '''
    if len(text) >= limit:
        if isinstance(suffix, str):
            # If we have a suffix, then reduce the text string length further by
            # the length of the suffix and then concatenate both the text and
            # suffix together.
            return '{}{}'.format(text[: limit - len(suffix)], suffix)
        else:
            # If no suffix, then simply reduce the string size.
            return text[:limit]
    return text


def check(name, obj, expected_type, **kwargs):
    '''
    Check function for validating that inputs we are receiving are of the right
    type, have the expected values, and can handle defaults as necessary.

    Args:
        name (str): The name of the object (for exception reporting)
        obj (obj): The object that we will be checking
        expected_type (type):
            The expected type of object that we will check against.
        choices (list, optional):
            if the object is only expected to have a finite number of values
            then we can check to make sure that our input is one of these
            values.
        default (obj, optional):
            if we want to return a default setting if the object is None,
            we can set one here.
        case (string, optional):
            if we want to force the object values to be upper or lower case,
            then we will want to set this to either ``upper`` or ``lower``
            depending on the desired outcome.  The returned object will then
            also be in the specified case.
        pattern (string, optional):
            Specify a regex pattern from the pattern map variable.
        regex (str, optional):
            Validate that the value of the object matches this pattern.
        items_type (type, optional):
            If the expected type is an iterable, and if all of the items
            within that iterable are expected to be a given type, then
            specifying the type here will enable checking each item within
            the iterable.  NOTE: this will traverse the iterable.

    Returns:
        :obj:`Object`:
            Either the object or the default object depending.

    Examples:
        Ensure that the value is an integer type:

        >>> check('example', val, int)

        Ensure that the value of val is within 0 and 100:

        >>> check('example', val, int, choices=list(range(100)))
    '''
    def validate_regex_pattern(regex, obj):
        if len(re.findall(regex, str(obj))) <= 0:
            raise UnexpectedValueError(
                '{} has value of {}.  Does not match pattern {}'.format(
                    name, obj, regex))

    def validate_choice_list(choices, obj):
        if obj not in choices:
            raise UnexpectedValueError(
                '{} has value of {}.  Expected one of {}'.format(
                    name, obj, ','.join([str(i) for i in choices])))

    def validate_expected_type(expected, obj):
        if not isinstance(obj, expected):
            raise TypeError('{} is of type {}.  Expected {}.'.format(
            name,
            obj.__class__.__name__,
            expected_type.__name__
                if hasattr(expected_type, '__name__') else expected_type
        ))

    def validate_normalized(obj, func, arg):
        if isinstance(obj, (list, tuple)):
            # If the object is a list or tuple type, then lets ensure that
            # all of the items within the obj .
            for item in obj:
                func(arg, item)
        else:
            func(arg, obj)

    pmap = dict_merge({
        'uuid': r'^[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}$',
        'email': r'^[a-zA-Z0-9_.+-]+@[a-zA-Z0-9-]+\.[a-zA-Z0-9-.]+$',
        'hex': r'^[a-fA-f0-9]+$',
        'url': r'^(https?:\/\/)?([\da-z\.-]+)\.([a-z\.]{2,6})([\/\w \.-]*)*\/?$',
        'ipv4': r'^([0-9]{1,3}\.){3}[0-9]{1,3}(\/([0-9]|[1-2][0-9]|3[0-2]))?$',
        'ipv6': r'(([0-9a-fA-F]{1,4}:){7,7}[0-9a-fA-F]{1,4}|([0-9a-fA-F]{1,4}:){1,7}:|([0-9a-fA-F]{1,4}:){1,6}:[0-9a-fA-F]{1,4}|([0-9a-fA-F]{1,4}:){1,5}(:[0-9a-fA-F]{1,4}){1,2}|([0-9a-fA-F]{1,4}:){1,4}(:[0-9a-fA-F]{1,4}){1,3}|([0-9a-fA-F]{1,4}:){1,3}(:[0-9a-fA-F]{1,4}){1,4}|([0-9a-fA-F]{1,4}:){1,2}(:[0-9a-fA-F]{1,4}){1,5}|[0-9a-fA-F]{1,4}:((:[0-9a-fA-F]{1,4}){1,6})|:((:[0-9a-fA-F]{1,4}){1,7}|:)|fe80:(:[0-9a-fA-F]{0,4}){0,4}%[0-9a-zA-Z]{1,}|::(ffff(:0{1,4}){0,1}:){0,1}((25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])\.){3,3}(25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])|([0-9a-fA-F]{1,4}:){1,4}:((25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])\.){3,3}(25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9]))',
    }, kwargs.get('pattern_map', dict()))


    # We have a simple function to convert the case of string values so that
    # we can ensure correct output.

    # Convert the case of the inputs.
    obj = force_case(obj, kwargs.get('case'))
    kwargs['choices'] = force_case(kwargs.get('choices'), kwargs.get('case'))
    kwargs['default'] = force_case(kwargs.get('default'), kwargs.get('case'))

    # If the object sent to us has a None value, then we will return None.
    # If a default was set, then we will return the default value.
    if obj == None:
        return kwargs.get('default')

    # If we are checking for a string type, we will also want to check for
    # unicode type transparently, so add the unicode type to the expected
    # types list.  NOTE this is for Python2 only, as Python3 treats all
    # strings as type string.
    if expected_type == str:
        try:
            expected_type = (str, unicode)
        except NameError:
            pass

    # If the object is none of the right types then we want to raise a
    # TypeError as it was something we weren't expecting.
    validate_expected_type(expected_type, obj)

    if kwargs.get('items_type'):
        # If the items within the list should also be of a specific type,
        # we can check those as well
        for item in obj:
            validate_expected_type(kwargs.get('items_type'), item)

    # if the object is only expected to have one of a finite set of values,
    # we should check against that and raise an exception if the the actual
    # value is outside of what we expect.
    if kwargs.get('choices'):
        validate_normalized(obj, validate_choice_list, kwargs.get('choices'))

    # If a pattern was specified, then we will want to pull the pattern from
    # the pattern map and validate that the
    if kwargs.get('pattern') and kwargs.get('pattern') in pmap.keys():
        validate_normalized(obj, validate_regex_pattern,
            pmap[kwargs.get('pattern')])

    # If there wasn't a pattern matching that identifier, then throw an
    # IndexError
    elif kwargs.get('pattern') and kwargs.get('pattern') not in pmap.keys():
        raise IndexError('pattern name {} not found in map'.format(pattern))

    # If a raw regex pattern was provided instead, then we will pass that over
    # and validate
    elif kwargs.get('regex'):
        validate_normalized(obj, validate_regex_pattern, kwargs.get('regex'))

    # if we made it this gauntlet without an exception being raised, then
    # assume everything is good to go and return the object passed to us
    # initially.
    return obj


'''
Endpoints
=========

.. autoclass:: APIEndpoint
    :members:
    :private-members:
'''


class APIEndpoint(object):
    '''
    APIEndpoint is the base model for which all API endpoint classes are
    sired from.  The main benefit is the addition of the ``_check()``
    function from which it's possible to check the type & content of a
    variable to ensure that we are passing good data to the API.

    Args:
        api (APISession):
            The APISession (or sired child) instance that the endpoint will
            be using to perform calls to the API.
    '''
    def __init__(self, api):
        self._api = api
        self._log = api._log


from setuptools import setup, find_packages


long_description = 'Please refer to https://restfly.readthedocs.io'

setup(
    name='restfly',
    version='1.1',
    description='A library to make API wrappers creation easier',
    author='Steve McGrath',
    long_description=long_description,
    author_email='steve@chigeek.com',
    url='https://github.com/stevemcgrath/restfly',
    license='MIT',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Topic :: Software Development',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Application Frameworks',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
    keywords='tenable tenable_io securitycenter containersecurity',
    packages=find_packages(exclude=['docs', 'tests']),
    install_requires=[
        'requests>=2.19',
    ],
)